/**
* @file Uzewav.c
*
* @brief uzewave library based on Uzeamp
*
* @version 1.1
*
* @date 2012 
*
* @author  Hartmut Wendt  info@hwhardsoft.de.vu	<http://www.hwhardsoft.de.vu/>
* @author  Alec Bourque	  <http://www.uzebox.org>
*
* @copyright GNU Public License V3
* 
* @note
*   This program is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   You should have received a copy of the GNU General Public License
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <stdbool.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <uzebox.h>
#include <mmc.h>
#include <avr/interrupt.h>
#include "uzewav.h"



/** 
 @brief array with file entries
*/
File files[MAX_WAVES];

/** 
 @brief queue with wave files for playing 
*/
char wav_queue[7] = {-99,-99,-99,-99,-99,-99,-99};

/** 
 @brief voice output of a file is running
*/
bool playing=false;
unsigned long sectorNo=0;
unsigned int cur,playingFile=0,x,y;
unsigned long songSize=0;

/** 
 @brief SD card detected
*/
bool cardDetected=false;
long bootRecordSector;
int reservedSectors;
int sectorsPerFat;
unsigned char maxRootDirectorySectors;
long dirTableSector;


uint32_t mmc_mixer_sector;
bool mmc_mixer_active=false;



/**
init the SD card for voice output
@param void
@retval 0  SD card init FAILED
@retval 1  FIRST READ FAILED 
@retval 2  SD card init ok
*/
u8 SD_init(void) {  

  unsigned char temp;
  int timeout=0;

   do { temp = mmc_init(sector.buffer);
  		timeout++;
		if(timeout>10){
			//  SD Card init FAILED
			return(0);		
		}
   	}while (temp);
	
	// SD Card init OK
	timeout=0;

	do {
		temp = mmc_readsector(0);
		timeout++;
		if(timeout>6000){
   			// FIRST READ FAILED 
			return (1);
		}
	}while (temp);
	return(2); 	
}


/**
convert the first 2 chars of WAV filename to a number
@param pointer to filename
@retval 0..98  numeric values of two first chars in filename
@retval 99  first 2 chars are not numeric values
*/
int convert_filename(unsigned char *fname) {

u8 u1;
int value = 0;

  for (u1 = 0; u1 < 2; u1++){

  	if (*(fname + u1) > '9')  return(100); // not a valid number
  	if (*(fname + u1) < '0')  return(100); // not a valid number
	value *= 10;
	value += *(fname + u1) - 0x30;	
  }
  return(value);
} 

/**
play automaticaly the wave files from queue
@param void
@return  none
*/
void play_queue(){
	u8 u1;
	if(playing) return;	
	for(u1=0; u1 < 7; u1++) {
		if (wav_queue[u1] != -99) {
			playingFile = wav_queue[u1];  			
			wav_queue[u1] = -99;
			startSong();
			return;
		}
	}
}

/**
start playing of voice file
@param void
@return  none
*/
void startSong(){
		if(playing) stopSong();
		sectorNo=files[playingFile].firstSector+1; //skip WAV header
		songSize=files[playingFile].fileSize/512;
		
		playing=true;
		mmc_mixerStart(sectorNo);		
}


/**
stop playing of voice file before the end of voice file
@param void
@return  none
*/
void stopSong(){
	mmc_mixerStop();
	playing=false;
}

/**
stop playing of voice file automaticaly at the end of voice file
@param void
@return  none
*/
void stop_auto(){	

    if(playing == false) return; 
	if(sectorNo>=(songSize+files[playingFile].firstSector)){
		mmc_mixerStop();
		WaitVsync(3);
		playing=false;
	}

}


long dirTableSector;
long sectorsPerCluster;
long maxRootDirectoryEntries;
long bytesPerSector;


/**
init the FAT of the SD card
@param pointer to buffer
@return  none
*/
void fat_init(unsigned char *buffer){

	//read MBR
	mmc_readsector(0);

	//read boot record
	bootRecordSector= ((MBR*)buffer)->partition1.startSector;
	mmc_readsector(bootRecordSector);

	reservedSectors=((BootRecord*)buffer)->reservedSectors;
	sectorsPerFat=((BootRecord*)buffer)->sectorsPerFat;
	sectorsPerCluster=((BootRecord*)buffer)->sectorsPerCluster;

	maxRootDirectoryEntries=((BootRecord*)buffer)->maxRootDirectoryEntries;
	maxRootDirectorySectors=((maxRootDirectoryEntries * 32)/BYTES_PER_SECTOR);
		

}


/**
get the file sector from the SD card
@param pointer to a directory table entry
@return  directory table
*/
long GetFileSector(DirectoryTableEntry *file){
	return dirTableSector+maxRootDirectorySectors+((file->firstCluster-2)*sectorsPerCluster);
}


/**
find valid files with extension of *char in the directory and store their first sector & filesize
@param pointer to string in flash with the file extension for example "GER" or "ENG"
@return  count of found files
*/
u8 find_files(const char *s){

	unsigned char c = 0, fileCount=0, page = 0;
	unsigned int i,j;
	
	if (!cardDetected) return(0);
	
	do{

		mmc_readsector(dirTableSector+page);

		//read all entries in the sector
		for(i=0;i<16;i++){
			if((sector.files[i].fileAttributes & (FAT_ATTR_HIDDEN|FAT_ATTR_SYSTEM|FAT_ATTR_VOLUME|FAT_ATTR_DIRECTORY|FAT_ATTR_DEVICE))==0){
				c=sector.files[i].filename[0];
				if(c == 0){

					break;
				}else if(c!=0xe5 && c!=0x05 && c!=0x2e &&
					sector.files[i].extension[0]==pgm_read_byte(&(s[0])) &&
					sector.files[i].extension[1]==pgm_read_byte(&(s[1])) &&
					sector.files[i].extension[2]==pgm_read_byte(&(s[2])) 
					){									

					//filesFirstSector[fileCount]=GetFileSector(&sector.files[i]);
					
					j = convert_filename(sector.files[i].filename);
					if (j < 100) {
						//Print(10,20,PSTR("GUELTIGES FILE"));
						//PrintInt(10,21,j,true);
						//PrintLong(15,22,sector.files[i].fileSize);
						
						files[j].fileSize=sector.files[i].fileSize;				
						files[j].firstSector=GetFileSector(&sector.files[i]);	
						fileCount++;
					/*
					} else {
						Print(10,20,PSTR("UNGUELTIGES FILE"));
						PrintRam(10,21,sector.files[i].filename);*/
					}				
					

					// WaitVsync(60);

					if(fileCount==MAX_WAVES) break;
				}				
			}
			
		}
		
		page++;

	}while(c!=0 && fileCount<MAX_WAVES && page<maxRootDirectorySectors);

	return(fileCount);
}


int mixerRead,sectorRead;

/**
start the mixer for sound output
@param LBA (Logical Block Addressing)
@return  none
*/
void mmc_mixerStart(uint32_t lba){
    if(mmc_mixer_active==true){
        mmc_send_command(12,0,0); //stop current transfer
        mmc_clock_and_release();
    }
       
    // send the multiple block read command and logical sector address
    mmc_mixer_sector=lba;
    mmc_send_command(18,(mmc_mixer_sector>>7) & 0xffff, (mmc_mixer_sector<<9) & 0xffff);   
    mixerRead=0;
    sectorRead=0;   
    mmc_mixer_active=true;
}

/**
stop the mixer for sound output
@param void
@return  none
*/
void mmc_mixerStop(){
    if(mmc_mixer_active==true){
        mmc_send_command(12,0,0); //stop transfers
        mmc_clock_and_release();

        //Initialize the mixer buffer
        for(int i=0;i<MIX_BANK_SIZE*2;i++){
            mix_buf[i]=0x80;
        }       
    }   
    mmc_mixer_active=false;
}

/**
get current sector from SD card for playing
@param void
@return  current sector
*/
uint32_t mmc_mixerGetCurrentSector(){
    return mmc_mixer_sector;
}


/**
sound output -
call once on each vsync
@param void
@return  none
*/
void mmc_processMixer()
{
   

    int retVal=0;
    uint8_t *buf=mix_buf;   


    if(mmc_mixer_active){

        if(mix_bank==1){
            buf+=MIX_BANK_SIZE;
        }


        do{
       
            if(sectorRead==0){               
                if (mmc_datatoken() != 0xfe)    // if no valid token
                {
                    mmc_clock_and_release();    // cleanup and   
                    Print(10,10,PSTR("TOKEN!"));
                    return;
                }
            }

            do{            // read sector data

                *buf++=spi_byte(0xff);
             
                mixerRead++;
                sectorRead++;

            }while(mixerRead<MIX_BANK_SIZE && sectorRead<512);
   

            if(sectorRead==512){
                spi_byte(0xff);                    // ignore dummy checksum
                spi_byte(0xff);                    // ignore dummy checksum
               
           
                mmc_mixer_sector++;
                sectorRead=0;
            }

            if(mixerRead==MIX_BANK_SIZE){
                mixerRead=0;
                retVal=1;       
            }


        }while(retVal==0);
    }
   

}
